# -*- Mode: Python; tab-width: 4 -*-

# TODO: completion of local symbols with the tab key.

import StringIO
import os
import string
import sys
import pprint
	
import completion

import mvc
import mvc_cell_win

import windc
import wingdi
import winwin

std_files = None, None, None

class python_interpreter_window (mvc_cell_win.text_edit_window):

	herald_color = (  0,   0, 192)
	input_color =  (  0,   0,   0)
	output_color = (  0, 128,   0)
	error_color =  (192,   0,   0)
	counter = 0

	def create (*args):
		self = apply (mvc_cell_win.text_edit_window.create, args)
		python_interpreter_window.counter = python_interpreter_window.counter + 1
		self.get_parent().set_window_text ('Python Interpreter %d' % self.counter)
		self.set_model (
			mvc.sequence_model (
				self.make_herald()
				)
			)
		return self

	def prepare_dc (self, dc):
		mvc_cell_win.text_edit_window.prepare_dc (self, dc)
		dc.select_object (self.font)
		dc.set_bk_color()
		dc.set_bk_mode (windc.TRANSPARENT)
		dc.set_text_align (wingdi.TA_UPDATECP)

	def make_herald (self):
		herald = (
			'Python DynWin Interpreter Copyright 1998 Nightmare Software',
			'Python %s on Win32' % (sys.version),
			sys.copyright,
			)
		result = []
		for line in herald:
			result.append ((self.herald_color, line))
		return result

	def prompt (self):
		self.model.append ((self.input_color, '>>> '))
		n = len (self.model)
		self.cursor_pos = (4, n-1)
		self.scoot()
		self.set_caret_pos()

	def scoot (self):
		"scroll the window up to make room for output"
		vm, hm = self.scroll_models
		cx, cy = self.get_cell_size()
		w, h = self.size
		n = len (self.model)

		lx, ly = self.get_limits()

		vm.limit = ly
		vm.page = h/cy

		if (h < (n * cy)):
			vm.position = ly - (h/cy)

		if len(self.model):
			hm.limit = lx
			hm.page = w/cx

	def get_limits (self):
		if len(self.model):
			lx, ly = (
				max (map (lambda x: len(x[1]), self.model)),
				len (self.model)
				)
		else:
			lx, ly = 0, 0
		return lx, ly

	def WM_LBUTTONDOWN (self, wparam, lparam):
		# don't let the user move the cursor position.
		return 0

	def WM_CHAR (self, wparam, lparam):

		ch = chr (wparam)
		w, h = self.size
		x, y = self.cursor_pos
		cx, cy = self.get_cell_size()

		color, line = self.model[y]

		if ch == '\r':
			# carriage return
			if line and line[:4] == '>>> ':
				self.handle_line_input (line[4:])
				self.line_history_index = 0
			else:
				self.prompt()

		elif ch == '\010':
			# backspace
			self.model[y] = color, line[:-1]
			self.cursor_pos = x-1, y

		elif ch == '\011':
			# tab, completion
			self.complete()
			
		elif ch in ('\020', '\016'):
			# ctrl-p previous line, ctrl-n next line
			if ch == '\020':
				delta = +1
			else:
				delta = -1
			self.line_history_index = (self.line_history_index + delta) % len(self.line_history)
			line = self.line_history[-self.line_history_index]
			self.model[y] = color, line
			self.cursor_pos = len(line), y
		else:
			chars = map (None, line)
			chars.insert (x, ch)
			self.model[y] = ((self.input_color, string.join (chars, '')))
			x = x + 1
			self.cursor_pos = x, y

		self.set_caret_pos()

		return 1

	def draw_line (self, dc, index):
		cx, cy = self.get_cell_size()
		y = cy * index
		dc.move_to ((0, y))
		color, line = self.model[index]
		dc.set_text_color (color)
		dc.text_out ((0,0), line)
		px, py = dc.move_to_ex ((0,0))

	multi_line = 0

	local_env = {}

	line_history = None
	line_history_index = 0

	def handle_line_input (self, line):
		# check for special case inputs...
		if not line and not self.multi_line:
			self.prompt()
			return
		else:
			# manage line history
			hline = '>>> ' + line
			if self.line_history is None:
				self.line_history = []

			if self.line_history and self.line_history[-1] == hline:
				# don't fill the history with copies of a single line
				pass
			else:
				self.line_history.append (hline)
				
		global std_files
		std_files = sys.stdin, sys.stdout, sys.stderr
		try:
			of = StringIO.StringIO()
			ef = StringIO.StringIO()
			sys.stdout = of
			sys.stderr = ef
			try:
				try:
					# first try eval
					co = compile (line, repr(self), 'eval')
					result = eval (co, self.local_env)
					if result is not None:
						pprint.pprint (result)
					self.local_env['_'] = result
				except SyntaxError:
					# now try exec
					co = compile (line, repr(self), 'exec')
					exec co in self.local_env
			except:
				self.write_traceback (
					sys.exc_type,
					sys.exc_value,
					sys.exc_traceback,
					sys.stderr
					)
		finally:
			sys.stdin, sys.stdout, sys.stderr = std_files

		ef.seek (0)
		lines = ef.readlines()
		self.append_lines (lines, self.error_color)
		of.seek (0)
		lines = of.readlines()
		self.append_lines (lines, self.output_color)
		self.prompt()


	complete_chars = string.letters + string.digits + '_.'

	def complete (self):
		# try to complete a partially-typed symbol...
		color, line = self.model[-1]

		i = len(line) - 1
		while i and line[i] in self.complete_chars:
			i = i - 1

		part = line[i+1:]

		if not part:
			winwin.user32.MessageBeep (0)
		else:
			# search local symbols
			prefix, matches = completion.completions (part, self.local_env)
			if len(matches) == 1:
				sym = matches[0]
				line = line[:i+1] + sym
				self.model[-1] = (color, line)
				x, y = self.cursor_pos
				self.cursor_pos = len(line), y
				self.set_caret_pos()
			else:
				line = line + prefix
				lines = ['completions: ' + string.join (matches, ', '), line]
				self.append_lines (lines, self.input_color)
				x, y = self.cursor_pos
				self.cursor_pos = len(line), y+2
				self.set_caret_pos()
				winwin.user32.MessageBeep (0)

	def append_lines (self, lines, color):
		n = len (self.model)
		r = []

		for line in lines:
			r.append ((color, self.frob_line (line)))

		self.model[n:n] = r
		self.scoot()
		
	def frob_line (self, line):

		while line and line[-1] in '\r\n':
			line = line[:-1]

		return string.join (string.split (line, '\t'), '    ')

	def flush_output (self):
		self.model[:] = []
		self.scoot()

	def write_traceback (self, t, v, tb, file):
		tbinfo = []
		file.write ('Traceback (innermost last)\n')
		file.write ('%s: %s\n' % (t,v))
		r = repr(self)
		while 1:
			filename = tb.tb_frame.f_code.co_filename
			if filename == r:
				filename = "<stdin>"
			file.write (
				'  File "%s", line %d in %s\n' % (
					filename,
					tb.tb_lineno,
					tb.tb_frame.f_code.co_name
					)
				)
			tb = tb.tb_next
			if not tb:
				break

class python_interpreter_component (mvc_cell_win.text_edit_component):

	sequence_view_class = python_interpreter_window

	def prompt (self):
		self.sequence_window.prompt()

	def flush_output (self):
		self.sequence_window.flush_output()

def new_window ():
	w = python_interpreter_component(
		style = winwin.WS_OVERLAPPEDWINDOW
		)
	w.create()
	w.show_window()
	w.prompt()
	return w

if __name__ == '__main__':
	new_window()
	import msgloop
	msgloop.go()
